class DeferredListObserver(object):
    """
    Call a list of observer functions with information about firing events
    that occur on a set of deferreds. Observers are called with event
    information in the order they are added (via C{observe}).

    @param maintainHistory: if C{True} a history of all events is maintained.
        This can be replayed to newly added observers and is accessible to
        class instances. If C{False}, the default, no history is kept.
    @ivar history: a C{list} of (index, success, value) tuples, in the order
        that the deferreds in the set fired (this will generally not be the
        order in which the deferreds are added to the set). The history
        attribute will only exist if C{maintainHistory} (above) is C{True}.
    @ivar successCount: the number of observed deferreds that have been called
        successfully.
    @ivar failureCount: the number of observed deferreds that have been
        errored.
    @ivar pendingCount: the number of observed deferreds that have not yet been
        called or errored.
    """

    def __init__(self, maintainHistory=False):
        self._maintainHistory = maintainHistory
        if maintainHistory:
            self.history = []
        self.successCount = self.failureCount = self.pendingCount = 0
        self._observers = []

    def _makeCallbacks(self, index):

        def callback(value):
            self.pendingCount -= 1
            self.successCount += 1
            event = (index, True, value)
            if self._maintainHistory:
                self.history.append(event)
            for observer in self._observers:
                observer(*event)
            return value

        def errback(value):
            self.pendingCount -= 1
            self.failureCount += 1
            event = (index, False, value)
            if self._maintainHistory:
                self.history.append(event)
            for observer in self._observers:
                observer(*event)
            return value

        return (callback, errback)

    def append(self, deferred):
        """
        Monitor a deferred.

        @param deferred: An instance of L{twisted.internet.defer.Deferred}.
        @return: the passed deferred.
        """
        index = self.successCount + self.failureCount + self.pendingCount
        callback, errback = self._makeCallbacks(index)
        self.pendingCount += 1
        return deferred.addCallbacks(callback, errback)

    def observe(self, observer, replayHistory=False):
        """
        Add an observer function that will be called (as below) with details
        of deferred firings.

        @param observer: a C{function} that will be called with 3 arguments
            each time one of the observed deferreds in the set fires. The
            arguments will be:
                - The index of the deferred that fired.
                - C{True} if the deferred was called, C{False} if it errored.
                - The value passed to the callback or errback.
        @param replayHistory: if C{True}, the history of deferred firings
            that occurred prior to this observer being added will be sent
            to the observer. If no history is being maintained, C{RuntimeError}
            will be raised.
        """
        if replayHistory:
            if self._maintainHistory:
                for event in self.history:
                    observer(*event)
            else:
                raise RuntimeError('Cannot replay non-existent event history '
                                   'to new observer')

        self._observers.append(observer)



#!/usr/bin/env python

# The description below is an RST version of the beginning of README.md
# (I'm not a fan of RST, but PyPI knows how to display it.)

description = """\
txdlo - a (Twisted) deferred list observer
==========================================

``txdlo`` is a Python package that provides a class called
``DeferredListObserver``.

As you might guess, ``DeferredListObserver`` lets you observe callback and
errback events from a list of `Twisted <http://twistedmatrix.com>`_
`deferreds
<http://twistedmatrix.com/documents/current/core/howto/defer.html>`_.  You
can add observers that will be passed information about deferreds firing.
You can add deferreds to the observed list at any time, which is very
useful if you're dynamically creating deferreds that you want to monitor.

The class can be used to easily build functions or classes that provide
deferreds that fire when arbitrary combinations of events from the observed
deferreds have occurred.

For example you can write functions or classes that support deferreds that

* Implement Twisted's ``DeferredList`` or simple variants of it, or that let
  you separate the various behaviors of ``DeferredList`` into simpler
  functions.
* Provide a deferred that fires when N of the observed deferreds have fired.
* Provide a deferred that ignores errors until one of the observed deferred
  succeeds, only firing with an error if all the observed deferreds fail.
* Or (a more involved example), suppose you have 3 methods that can return
  you a user's avatar: a fast local cache, a filesystem, and a slow network
  call to Gravatar. You want to write a deferred-returning function that
  launches all three lookups at once and fires its deferred with the first
  answer. But if the cache and/or filesystems fails first, you don't want
  to trigger an error, you instead want to take the result from Gravatar
  and add it to the cache and/or filesystem, as well firing the returned
  deferred with the result (wherever it comes from). Only if all three
  lookups fail do you want to errback the deferred you returned.

The source for ``txdlo`` lives at `https://github.com/terrycojones/txdlo
<https://github.com/terrycojones/txdlo>`_.

"""

d = dict(name='txdlo',
         version='0.1.8',
         provides=['txdlo'],
         maintainer='Terry Jones',
         maintainer_email='terry@jon.es',
         url='https://github.com/terrycojones/txdlo',
         download_url='https://github.com/terrycojones/txdlo',
         packages=['txdlo'],
         include_package_data=True,
         keywords=['twisted deferred observer'],
         classifiers=[
             'Programming Language :: Python',
             'Framework :: Twisted',
             'Development Status :: 4 - Beta',
             'Intended Audience :: Developers',
             'License :: OSI Approved :: Apache Software License',
             'Operating System :: OS Independent',
             'Topic :: Software Development :: Libraries :: Python Modules',
         ],
         description='A Twisted class for observing a set of deferreds.',
         long_description=description)

try:
    from setuptools import setup
    _ = setup  # Keeps pyflakes from complaining.
except ImportError:
    from distutils.core import setup
else:
    d['install_requires'] = ['Twisted']

setup(**d)
